/*============================================================================
  File:     GroupingSetsPerformance.sql

  Summary:  This script shows some of the performance gains when using GROUPING
			SETS vs multiple GROUP BY clauses and UNION query.
			
			Run these scripts with "Include actual execution plan" ON. Note the 
			Rollup Information property on the aggregate iterator farthest to the
			left in the graphic execution plan. NOTE: Click on the iterator and
			use the Properties Window to see the rollup information property.
			
			The last example shows that to disambiguate between NULL meaning 
			rollup level and a true NULL database value, you must use the GROUPING
			function. 

  Date:     August 2008

  SQL Server Version: 10.0.1600.22 (RTM)
------------------------------------------------------------------------------
  Written by Bob Beauchemin, SQLskills.com

  For more scripts and sample code, check out http://www.SQLskills.com

  This script is intended only as a supplement to SQL Server Jumpstart program.
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

use tempdb
go

create table T (
 a int,
 b int,
 c int,
 x int
)
go

insert T values(1,2,3,8)
insert T values(1,2,4,1)
insert T values(1,2,5,5)
insert T values(1,3,4,9)
go

select a, b, c, sum(x) as sumx from T
group by grouping sets (a, (b), (a, b, c))
go

-- this is equivalent to the previous query
-- but the query plan is worse (60-40)
select a, NULL as b, NULL as c, sum(x) as sumx from T
group by a
union all
select NULL as a, b, NULL as c, sum(x) as sumx from T
group by b
union all
select a, b, c, sum(x) as sumx from T
group by a, b, c
go

select a, b, c, sum(x) as sumx, grouping_id(a,b,c) as gid,
  GROUPING(a) as [GroupingByA], GROUPING(b) as [GroupingByB], GROUPING(c) as [GroupingByC]
from T
group by grouping sets (a, (b), (a, b, c))
go

-- rollup plan
SELECT a, b, c, sum(x) FROM T
GROUP BY a ,b ,c 
WITH ROLLUP
GO

-- cube plan
SELECT a, b, c, sum(x) FROM T
GROUP BY a ,b ,c 
WITH CUBE
GO

-- cube plan without base aggregation or spool
SELECT a, b, c  FROM T
GROUP BY GROUPING SETS ( (), (a), (a,b), (b,c), (c), (a,c), (b) )
GO

--Issue with nulls

-- insert a NULL value
insert T values(NULL,10,10,10)
go

--Difficult to distinguish between different levels of grouping, show multiple rows with the same values for a, b and c
--need to use GROUPING function
SELECT a, b, c, sum(x) , GROUPING(a) as [GroupingByA], GROUPING(b) as [GroupingByB], GROUPING(c) as [GroupingByC]
FROM T
GROUP BY GROUPING SETS ((a,b,c), (a), ())
ORDER BY 1,2,3
go


